==> jooby-flyway-1.3.0-sources.jar: org/jooby/flyway/Flywaydb.java <==
/**
 * Apache License
 * Version 2.0, January 2004
 * http://www.apache.org/licenses/
 *
 * TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 * 1. Definitions.
 *
 * "License" shall mean the terms and conditions for use, reproduction,
 * and distribution as defined by Sections 1 through 9 of this document.
 *
 * "Licensor" shall mean the copyright owner or entity authorized by
 * the copyright owner that is granting the License.
 *
 * "Legal Entity" shall mean the union of the acting entity and all
 * other entities that control, are controlled by, or are under common
 * control with that entity. For the purposes of this definition,
 * "control" means (i) the power, direct or indirect, to cause the
 * direction or management of such entity, whether by contract or
 * otherwise, or (ii) ownership of fifty percent (50%) or more of the
 * outstanding shares, or (iii) beneficial ownership of such entity.
 *
 * "You" (or "Your") shall mean an individual or Legal Entity
 * exercising permissions granted by this License.
 *
 * "Source" form shall mean the preferred form for making modifications,
 * including but not limited to software source code, documentation
 * source, and configuration files.
 *
 * "Object" form shall mean any form resulting from mechanical
 * transformation or translation of a Source form, including but
 * not limited to compiled object code, generated documentation,
 * and conversions to other media types.
 *
 * "Work" shall mean the work of authorship, whether in Source or
 * Object form, made available under the License, as indicated by a
 * copyright notice that is included in or attached to the work
 * (an example is provided in the Appendix below).
 *
 * "Derivative Works" shall mean any work, whether in Source or Object
 * form, that is based on (or derived from) the Work and for which the
 * editorial revisions, annotations, elaborations, or other modifications
 * represent, as a whole, an original work of authorship. For the purposes
 * of this License, Derivative Works shall not include works that remain
 * separable from, or merely link (or bind by name) to the interfaces of,
 * the Work and Derivative Works thereof.
 *
 * "Contribution" shall mean any work of authorship, including
 * the original version of the Work and any modifications or additions
 * to that Work or Derivative Works thereof, that is intentionally
 * submitted to Licensor for inclusion in the Work by the copyright owner
 * or by an individual or Legal Entity authorized to submit on behalf of
 * the copyright owner. For the purposes of this definition, "submitted"
 * means any form of electronic, verbal, or written communication sent
 * to the Licensor or its representatives, including but not limited to
 * communication on electronic mailing lists, source code control systems,
 * and issue tracking systems that are managed by, or on behalf of, the
 * Licensor for the purpose of discussing and improving the Work, but
 * excluding communication that is conspicuously marked or otherwise
 * designated in writing by the copyright owner as "Not a Contribution."
 *
 * "Contributor" shall mean Licensor and any individual or Legal Entity
 * on behalf of whom a Contribution has been received by Licensor and
 * subsequently incorporated within the Work.
 *
 * 2. Grant of Copyright License. Subject to the terms and conditions of
 * this License, each Contributor hereby grants to You a perpetual,
 * worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 * copyright license to reproduce, prepare Derivative Works of,
 * publicly display, publicly perform, sublicense, and distribute the
 * Work and such Derivative Works in Source or Object form.
 *
 * 3. Grant of Patent License. Subject to the terms and conditions of
 * this License, each Contributor hereby grants to You a perpetual,
 * worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 * (except as stated in this section) patent license to make, have made,
 * use, offer to sell, sell, import, and otherwise transfer the Work,
 * where such license applies only to those patent claims licensable
 * by such Contributor that are necessarily infringed by their
 * Contribution(s) alone or by combination of their Contribution(s)
 * with the Work to which such Contribution(s) was submitted. If You
 * institute patent litigation against any entity (including a
 * cross-claim or counterclaim in a lawsuit) alleging that the Work
 * or a Contribution incorporated within the Work constitutes direct
 * or contributory patent infringement, then any patent licenses
 * granted to You under this License for that Work shall terminate
 * as of the date such litigation is filed.
 *
 * 4. Redistribution. You may reproduce and distribute copies of the
 * Work or Derivative Works thereof in any medium, with or without
 * modifications, and in Source or Object form, provided that You
 * meet the following conditions:
 *
 * (a) You must give any other recipients of the Work or
 * Derivative Works a copy of this License; and
 *
 * (b) You must cause any modified files to carry prominent notices
 * stating that You changed the files; and
 *
 * (c) You must retain, in the Source form of any Derivative Works
 * that You distribute, all copyright, patent, trademark, and
 * attribution notices from the Source form of the Work,
 * excluding those notices that do not pertain to any part of
 * the Derivative Works; and
 *
 * (d) If the Work includes a "NOTICE" text file as part of its
 * distribution, then any Derivative Works that You distribute must
 * include a readable copy of the attribution notices contained
 * within such NOTICE file, excluding those notices that do not
 * pertain to any part of the Derivative Works, in at least one
 * of the following places: within a NOTICE text file distributed
 * as part of the Derivative Works; within the Source form or
 * documentation, if provided along with the Derivative Works; or,
 * within a display generated by the Derivative Works, if and
 * wherever such third-party notices normally appear. The contents
 * of the NOTICE file are for informational purposes only and
 * do not modify the License. You may add Your own attribution
 * notices within Derivative Works that You distribute, alongside
 * or as an addendum to the NOTICE text from the Work, provided
 * that such additional attribution notices cannot be construed
 * as modifying the License.
 *
 * You may add Your own copyright statement to Your modifications and
 * may provide additional or different license terms and conditions
 * for use, reproduction, or distribution of Your modifications, or
 * for any such Derivative Works as a whole, provided Your use,
 * reproduction, and distribution of the Work otherwise complies with
 * the conditions stated in this License.
 *
 * 5. Submission of Contributions. Unless You explicitly state otherwise,
 * any Contribution intentionally submitted for inclusion in the Work
 * by You to the Licensor shall be under the terms and conditions of
 * this License, without any additional terms or conditions.
 * Notwithstanding the above, nothing herein shall supersede or modify
 * the terms of any separate license agreement you may have executed
 * with Licensor regarding such Contributions.
 *
 * 6. Trademarks. This License does not grant permission to use the trade
 * names, trademarks, service marks, or product names of the Licensor,
 * except as required for reasonable and customary use in describing the
 * origin of the Work and reproducing the content of the NOTICE file.
 *
 * 7. Disclaimer of Warranty. Unless required by applicable law or
 * agreed to in writing, Licensor provides the Work (and each
 * Contributor provides its Contributions) on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 * implied, including, without limitation, any warranties or conditions
 * of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 * PARTICULAR PURPOSE. You are solely responsible for determining the
 * appropriateness of using or redistributing the Work and assume any
 * risks associated with Your exercise of permissions under this License.
 *
 * 8. Limitation of Liability. In no event and under no legal theory,
 * whether in tort (including negligence), contract, or otherwise,
 * unless required by applicable law (such as deliberate and grossly
 * negligent acts) or agreed to in writing, shall any Contributor be
 * liable to You for damages, including any direct, indirect, special,
 * incidental, or consequential damages of any character arising as a
 * result of this License or out of the use or inability to use the
 * Work (including but not limited to damages for loss of goodwill,
 * work stoppage, computer failure or malfunction, or any and all
 * other commercial damages or losses), even if such Contributor
 * has been advised of the possibility of such damages.
 *
 * 9. Accepting Warranty or Additional Liability. While redistributing
 * the Work or Derivative Works thereof, You may choose to offer,
 * and charge a fee for, acceptance of support, warranty, indemnity,
 * or other liability obligations and/or rights consistent with this
 * License. However, in accepting such obligations, You may act only
 * on Your own behalf and on Your sole responsibility, not on behalf
 * of any other Contributor, and only if You agree to indemnify,
 * defend, and hold each Contributor harmless for any liability
 * incurred by, or claims asserted against, such Contributor by reason
 * of your accepting any such warranty or additional liability.
 *
 * END OF TERMS AND CONDITIONS
 *
 * APPENDIX: How to apply the Apache License to your work.
 *
 * To apply the Apache License to your work, attach the following
 * boilerplate notice, with the fields enclosed by brackets "{}"
 * replaced with your own identifying information. (Don't include
 * the brackets!)  The text should be enclosed in the appropriate
 * comment syntax for the file format. We also recommend that a
 * file or class name and description of purpose be included on the
 * same "printed page" as the copyright notice for easier
 * identification within third-party archives.
 *
 * Copyright 2014 Edgar Espina
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jooby.flyway;

import com.google.inject.Binder;
import com.google.inject.Key;
import com.google.inject.name.Names;
import com.typesafe.config.Config;
import com.typesafe.config.ConfigFactory;
import static com.typesafe.config.ConfigFactory.empty;
import com.typesafe.config.ConfigValueType;
import static java.util.Objects.requireNonNull;
import org.flywaydb.core.Flyway;
import org.jooby.Env;
import org.jooby.Jooby.Module;
import org.jooby.funzy.Try;

import javax.sql.DataSource;
import java.util.ArrayList;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Properties;
import java.util.stream.Collectors;

/**
 * <h1>flyway module</h1>
 * <p>
 * Run database migrations on startup and exposes a {@link Flyway} instance.
 * </p>
 *
 * <p>This module depends on {@link org.jooby.jdbc.Jdbc} module.</p>
 *
 * <h2>usage</h2>
 *
 * <pre>
 * {
 *   use(new Jdbc());
 *
 *   use(new Flywaydb());
 * }
 * </pre>
 *
 * <p>
 * If for any reason you need to maintain two or more databases:
 * </p>
 *
 * <pre>
 * {
 *   use(new Jdbc("db1"));
 *   use(new Flywaydb("db1"));
 *
 *   use(new Jdbc("db2"));
 *   use(new Flywaydb("db2"));
 * }
 * </pre>
 *
 * <h2>migration scripts</h2>
 * <p>
 * {@link Flyway} looks for migration scripts at the <code>db/migration</code> classpath location.
 * We recommend to use <a href="http://semver.org">Semantic versioning</a> for naming the migration
 * scripts:
 * </p>
 *
 * <pre>
 * v0.1.0_My_description.sql
 * v0.1.1_My_small_change.sql
 * </pre>
 *
 * <h2>commands</h2>
 * <p>
 * It is possible to run {@link Flyway} commands on startup, default command is:
 * <code>migrate</code>.
 * </p>
 * <p>
 * If you need to run multiple commands, set the <code>flyway.run</code> property:
 * </p>
 *
 * <pre>
 * flyway.run = [clean, migrate, validate, info]
 * </pre>
 *
 * <h2>configuration</h2>
 * <p>
 * Configuration is done via <code>application.conf</code> under the <code>flyway.*</code> path.
 * There are some defaults setting that you can see in the appendix.
 * </p>
 *
 * @author edgar
 * @since 0.8.0
 */
public class Flywaydb implements Module {

  private enum Command {
    migrate {
      @Override
      public void run(final Flyway flyway) {
        flyway.migrate();
      }
    },

    clean {
      @Override
      public void run(final Flyway flyway) {
        flyway.clean();
      }
    },

    info {
      @Override
      public void run(final Flyway flyway) {
        flyway.info();
      }
    },

    validate {
      @Override
      public void run(final Flyway flyway) {
        flyway.validate();
      }
    },

    baseline {
      @Override
      public void run(final Flyway flyway) {
        flyway.baseline();
      }
    },

    repair {
      @Override
      public void run(final Flyway flyway) {
        flyway.repair();
      }
    };

    public abstract void run(Flyway flyway);
  }

  private String name;

  /**
   * Creates a new {@link Flywaydb}. The given name is use it to read flyway properties, keep in
   * mind that custom configuration will inherited all the properties from <code>flyway.*</code>.
   *
   * @param name Name of the property with flyway properties.
   */
  public Flywaydb(final String name) {
    this.name = requireNonNull(name, "Flyway name is required.");
  }

  /**
   * Creates a new {@link Flywaydb} module, using <code>db</code> as property.
   */
  public Flywaydb() {
    this("db");
  }

  @Override
  public void configure(final Env env, final Config conf, final Binder binder) {
    Config $base = flyway(conf.getConfig("flyway"));
    Config $flyway = Try.apply(() -> conf.getConfig(name).withFallback($base))
        .orElse($base);

    Flyway flyway = new Flyway();
    flyway.configure(props($flyway));
    if (!$flyway.hasPath("url")) {
      Key<DataSource> dskey = Key.get(DataSource.class, Names.named(name));
      DataSource dataSource = env.get(dskey)
          .orElseThrow(() -> new NoSuchElementException("DataSource missing: " + dskey));
      flyway.setDataSource(dataSource);
    }
    // bind
    env.serviceKey()
        .generate(Flyway.class, name, key -> binder.bind(key).toInstance(flyway));
    // run
    Iterable<Command> cmds = commands($flyway);
    // eager initialization
    cmds.forEach(cmd -> cmd.run(flyway));
  }

  private Config flyway(Config conf) {
    Config flyway = conf.root().entrySet().stream()
        .filter(it -> it.getValue().valueType() != ConfigValueType.OBJECT)
        .reduce(empty(), (seed, entry) -> seed.withValue(entry.getKey(), entry.getValue()),
            Config::withFallback);
    return flyway;
  }

  @Override
  public Config config() {
    return ConfigFactory.parseResources(getClass(), "flyway.conf");
  }

  @SuppressWarnings({"unchecked", "rawtypes"})
  private static Properties props(final Config config) {
    Properties props = new Properties();
    config.withoutPath("run").entrySet().forEach(prop -> {
      Object value = prop.getValue().unwrapped();
      if (value instanceof List) {
        value = ((List) value).stream().collect(Collectors.joining(","));
      }
      props.setProperty("flyway." + prop.getKey(), value.toString());
    });
    return props;
  }

  @SuppressWarnings("unchecked")
  private static Iterable<Command> commands(final Config config) {
    Object value = config.getAnyRef("run");
    List<String> commands = new ArrayList<>();
    if (value instanceof List) {
      commands.addAll((List<? extends String>) value);
    } else {
      commands.add(value.toString());
    }
    return commands.stream()
        .map(command -> Command.valueOf(command.toLowerCase()))
        .collect(Collectors.toList());
  }
}

==> jooby-flyway-1.4.0-sources.jar: org/jooby/flyway/Flywaydb.java <==
/**
 * Apache License
 * Version 2.0, January 2004
 * http://www.apache.org/licenses/
 *
 * TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 * 1. Definitions.
 *
 * "License" shall mean the terms and conditions for use, reproduction,
 * and distribution as defined by Sections 1 through 9 of this document.
 *
 * "Licensor" shall mean the copyright owner or entity authorized by
 * the copyright owner that is granting the License.
 *
 * "Legal Entity" shall mean the union of the acting entity and all
 * other entities that control, are controlled by, or are under common
 * control with that entity. For the purposes of this definition,
 * "control" means (i) the power, direct or indirect, to cause the
 * direction or management of such entity, whether by contract or
 * otherwise, or (ii) ownership of fifty percent (50%) or more of the
 * outstanding shares, or (iii) beneficial ownership of such entity.
 *
 * "You" (or "Your") shall mean an individual or Legal Entity
 * exercising permissions granted by this License.
 *
 * "Source" form shall mean the preferred form for making modifications,
 * including but not limited to software source code, documentation
 * source, and configuration files.
 *
 * "Object" form shall mean any form resulting from mechanical
 * transformation or translation of a Source form, including but
 * not limited to compiled object code, generated documentation,
 * and conversions to other media types.
 *
 * "Work" shall mean the work of authorship, whether in Source or
 * Object form, made available under the License, as indicated by a
 * copyright notice that is included in or attached to the work
 * (an example is provided in the Appendix below).
 *
 * "Derivative Works" shall mean any work, whether in Source or Object
 * form, that is based on (or derived from) the Work and for which the
 * editorial revisions, annotations, elaborations, or other modifications
 * represent, as a whole, an original work of authorship. For the purposes
 * of this License, Derivative Works shall not include works that remain
 * separable from, or merely link (or bind by name) to the interfaces of,
 * the Work and Derivative Works thereof.
 *
 * "Contribution" shall mean any work of authorship, including
 * the original version of the Work and any modifications or additions
 * to that Work or Derivative Works thereof, that is intentionally
 * submitted to Licensor for inclusion in the Work by the copyright owner
 * or by an individual or Legal Entity authorized to submit on behalf of
 * the copyright owner. For the purposes of this definition, "submitted"
 * means any form of electronic, verbal, or written communication sent
 * to the Licensor or its representatives, including but not limited to
 * communication on electronic mailing lists, source code control systems,
 * and issue tracking systems that are managed by, or on behalf of, the
 * Licensor for the purpose of discussing and improving the Work, but
 * excluding communication that is conspicuously marked or otherwise
 * designated in writing by the copyright owner as "Not a Contribution."
 *
 * "Contributor" shall mean Licensor and any individual or Legal Entity
 * on behalf of whom a Contribution has been received by Licensor and
 * subsequently incorporated within the Work.
 *
 * 2. Grant of Copyright License. Subject to the terms and conditions of
 * this License, each Contributor hereby grants to You a perpetual,
 * worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 * copyright license to reproduce, prepare Derivative Works of,
 * publicly display, publicly perform, sublicense, and distribute the
 * Work and such Derivative Works in Source or Object form.
 *
 * 3. Grant of Patent License. Subject to the terms and conditions of
 * this License, each Contributor hereby grants to You a perpetual,
 * worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 * (except as stated in this section) patent license to make, have made,
 * use, offer to sell, sell, import, and otherwise transfer the Work,
 * where such license applies only to those patent claims licensable
 * by such Contributor that are necessarily infringed by their
 * Contribution(s) alone or by combination of their Contribution(s)
 * with the Work to which such Contribution(s) was submitted. If You
 * institute patent litigation against any entity (including a
 * cross-claim or counterclaim in a lawsuit) alleging that the Work
 * or a Contribution incorporated within the Work constitutes direct
 * or contributory patent infringement, then any patent licenses
 * granted to You under this License for that Work shall terminate
 * as of the date such litigation is filed.
 *
 * 4. Redistribution. You may reproduce and distribute copies of the
 * Work or Derivative Works thereof in any medium, with or without
 * modifications, and in Source or Object form, provided that You
 * meet the following conditions:
 *
 * (a) You must give any other recipients of the Work or
 * Derivative Works a copy of this License; and
 *
 * (b) You must cause any modified files to carry prominent notices
 * stating that You changed the files; and
 *
 * (c) You must retain, in the Source form of any Derivative Works
 * that You distribute, all copyright, patent, trademark, and
 * attribution notices from the Source form of the Work,
 * excluding those notices that do not pertain to any part of
 * the Derivative Works; and
 *
 * (d) If the Work includes a "NOTICE" text file as part of its
 * distribution, then any Derivative Works that You distribute must
 * include a readable copy of the attribution notices contained
 * within such NOTICE file, excluding those notices that do not
 * pertain to any part of the Derivative Works, in at least one
 * of the following places: within a NOTICE text file distributed
 * as part of the Derivative Works; within the Source form or
 * documentation, if provided along with the Derivative Works; or,
 * within a display generated by the Derivative Works, if and
 * wherever such third-party notices normally appear. The contents
 * of the NOTICE file are for informational purposes only and
 * do not modify the License. You may add Your own attribution
 * notices within Derivative Works that You distribute, alongside
 * or as an addendum to the NOTICE text from the Work, provided
 * that such additional attribution notices cannot be construed
 * as modifying the License.
 *
 * You may add Your own copyright statement to Your modifications and
 * may provide additional or different license terms and conditions
 * for use, reproduction, or distribution of Your modifications, or
 * for any such Derivative Works as a whole, provided Your use,
 * reproduction, and distribution of the Work otherwise complies with
 * the conditions stated in this License.
 *
 * 5. Submission of Contributions. Unless You explicitly state otherwise,
 * any Contribution intentionally submitted for inclusion in the Work
 * by You to the Licensor shall be under the terms and conditions of
 * this License, without any additional terms or conditions.
 * Notwithstanding the above, nothing herein shall supersede or modify
 * the terms of any separate license agreement you may have executed
 * with Licensor regarding such Contributions.
 *
 * 6. Trademarks. This License does not grant permission to use the trade
 * names, trademarks, service marks, or product names of the Licensor,
 * except as required for reasonable and customary use in describing the
 * origin of the Work and reproducing the content of the NOTICE file.
 *
 * 7. Disclaimer of Warranty. Unless required by applicable law or
 * agreed to in writing, Licensor provides the Work (and each
 * Contributor provides its Contributions) on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 * implied, including, without limitation, any warranties or conditions
 * of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 * PARTICULAR PURPOSE. You are solely responsible for determining the
 * appropriateness of using or redistributing the Work and assume any
 * risks associated with Your exercise of permissions under this License.
 *
 * 8. Limitation of Liability. In no event and under no legal theory,
 * whether in tort (including negligence), contract, or otherwise,
 * unless required by applicable law (such as deliberate and grossly
 * negligent acts) or agreed to in writing, shall any Contributor be
 * liable to You for damages, including any direct, indirect, special,
 * incidental, or consequential damages of any character arising as a
 * result of this License or out of the use or inability to use the
 * Work (including but not limited to damages for loss of goodwill,
 * work stoppage, computer failure or malfunction, or any and all
 * other commercial damages or losses), even if such Contributor
 * has been advised of the possibility of such damages.
 *
 * 9. Accepting Warranty or Additional Liability. While redistributing
 * the Work or Derivative Works thereof, You may choose to offer,
 * and charge a fee for, acceptance of support, warranty, indemnity,
 * or other liability obligations and/or rights consistent with this
 * License. However, in accepting such obligations, You may act only
 * on Your own behalf and on Your sole responsibility, not on behalf
 * of any other Contributor, and only if You agree to indemnify,
 * defend, and hold each Contributor harmless for any liability
 * incurred by, or claims asserted against, such Contributor by reason
 * of your accepting any such warranty or additional liability.
 *
 * END OF TERMS AND CONDITIONS
 *
 * APPENDIX: How to apply the Apache License to your work.
 *
 * To apply the Apache License to your work, attach the following
 * boilerplate notice, with the fields enclosed by brackets "{}"
 * replaced with your own identifying information. (Don't include
 * the brackets!)  The text should be enclosed in the appropriate
 * comment syntax for the file format. We also recommend that a
 * file or class name and description of purpose be included on the
 * same "printed page" as the copyright notice for easier
 * identification within third-party archives.
 *
 * Copyright 2014 Edgar Espina
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jooby.flyway;

import com.google.common.collect.ImmutableSet;
import com.google.inject.Binder;
import com.google.inject.Key;
import com.google.inject.name.Names;
import com.typesafe.config.Config;
import com.typesafe.config.ConfigFactory;
import static com.typesafe.config.ConfigFactory.empty;
import com.typesafe.config.ConfigValueType;
import static java.util.Objects.requireNonNull;
import org.flywaydb.core.Flyway;
import org.jooby.Env;
import org.jooby.Jooby.Module;
import org.jooby.funzy.Try;

import javax.sql.DataSource;
import java.util.ArrayList;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Properties;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * <h1>flyway module</h1>
 * <p>
 * Run database migrations on startup and exposes a {@link Flyway} instance.
 * </p>
 *
 * <p>This module depends on {@link org.jooby.jdbc.Jdbc} module.</p>
 *
 * <h2>usage</h2>
 *
 * <pre>
 * {
 *   use(new Jdbc());
 *
 *   use(new Flywaydb());
 * }
 * </pre>
 *
 * <p>
 * If for any reason you need to maintain two or more databases:
 * </p>
 *
 * <pre>
 * {
 *   use(new Jdbc("db1"));
 *   use(new Flywaydb("db1"));
 *
 *   use(new Jdbc("db2"));
 *   use(new Flywaydb("db2"));
 * }
 * </pre>
 *
 * <h2>migration scripts</h2>
 * <p>
 * {@link Flyway} looks for migration scripts at the <code>db/migration</code> classpath location.
 * We recommend to use <a href="http://semver.org">Semantic versioning</a> for naming the migration
 * scripts:
 * </p>
 *
 * <pre>
 * v0.1.0_My_description.sql
 * v0.1.1_My_small_change.sql
 * </pre>
 *
 * <h2>commands</h2>
 * <p>
 * It is possible to run {@link Flyway} commands on startup, default command is:
 * <code>migrate</code>.
 * </p>
 * <p>
 * If you need to run multiple commands, set the <code>flyway.run</code> property:
 * </p>
 *
 * <pre>
 * flyway.run = [clean, migrate, validate, info]
 * </pre>
 *
 * <h2>configuration</h2>
 * <p>
 * Configuration is done via <code>application.conf</code> under the <code>flyway.*</code> path.
 * There are some defaults setting that you can see in the appendix.
 * </p>
 *
 * @author edgar
 * @since 0.8.0
 */
public class Flywaydb implements Module {

  private enum Command {
    migrate {
      @Override
      public void run(final Flyway flyway) {
        flyway.migrate();
      }
    },

    clean {
      @Override
      public void run(final Flyway flyway) {
        flyway.clean();
      }
    },

    info {
      @Override
      public void run(final Flyway flyway) {
        flyway.info();
      }
    },

    validate {
      @Override
      public void run(final Flyway flyway) {
        flyway.validate();
      }
    },

    baseline {
      @Override
      public void run(final Flyway flyway) {
        flyway.baseline();
      }
    },

    repair {
      @Override
      public void run(final Flyway flyway) {
        flyway.repair();
      }
    };

    public abstract void run(Flyway flyway);
  }

  private String name;

  /**
   * Creates a new {@link Flywaydb}. The given name is use it to read flyway properties, keep in
   * mind that custom configuration will inherited all the properties from <code>flyway.*</code>.
   *
   * @param name Name of the property with flyway properties.
   */
  public Flywaydb(final String name) {
    this.name = requireNonNull(name, "Flyway name is required.");
  }

  /**
   * Creates a new {@link Flywaydb} module, using <code>db</code> as property.
   */
  public Flywaydb() {
    this("db");
  }

  @Override
  public void configure(final Env env, final Config conf, final Binder binder) {
    Config $base = flyway(conf.getConfig("flyway"));
    Config $flyway = Try.apply(() -> conf.getConfig(name).withFallback($base))
        .orElse($base);

    Flyway flyway = new Flyway();
    Properties props = props($flyway);
    flyway.configure(props);
    if (!props.containsKey("flyway.url")) {
      Key<DataSource> dskey = Key.get(DataSource.class, Names.named(name));
      DataSource dataSource = env.get(dskey)
          .orElseThrow(() -> new NoSuchElementException("DataSource missing: " + dskey));
      flyway.setDataSource(dataSource);
    }
    // bind
    env.serviceKey()
        .generate(Flyway.class, name, key -> binder.bind(key).toInstance(flyway));
    // commands:
    Iterable<Command> cmds = commands($flyway);

    // eager initialization
    cmds.forEach(cmd -> cmd.run(flyway));
  }

  private Config flyway(Config conf) {
    Config flyway = conf.root().entrySet().stream()
        .filter(it -> it.getValue().valueType() != ConfigValueType.OBJECT)
        .reduce(empty(), (seed, entry) -> seed.withValue(entry.getKey(), entry.getValue()),
            Config::withFallback);
    return flyway;
  }

  @Override
  public Config config() {
    return ConfigFactory.parseResources(getClass(), "flyway.conf");
  }

  @SuppressWarnings({"unchecked", "rawtypes"})
  private static Properties props(final Config config) {
    Properties props = new Properties();
    config.withoutPath("run").entrySet().forEach(prop -> {
      Object value = prop.getValue().unwrapped();
      if (value instanceof List) {
        value = ((List) value).stream().collect(Collectors.joining(","));
      }
      String propertyName = prop.getKey();
      if (isFlywayProperty(propertyName)) {
        props.setProperty("flyway." + prop.getKey(), value.toString());
      }
    });
    return props;
  }

  /**
   * Filter residual/external properties from jdbc module. See #1044.
   * @param name Property name.
   * @return True if this is a valid flyway property.
   */
  static boolean isFlywayProperty(String name) {
    String setter = "set" + name;
    return Stream.of(Flyway.class.getMethods())
        .filter(m -> m.getName().equalsIgnoreCase(setter))
        .findFirst()
        .map(m -> true)
        .orElseGet(() -> ImmutableSet.of("driver", "url", "user", "password").contains(name));
  }

  @SuppressWarnings("unchecked")
  private static Iterable<Command> commands(final Config config) {
    Object value = config.getAnyRef("run");
    List<String> commands = new ArrayList<>();
    if (value instanceof List) {
      commands.addAll((List<? extends String>) value);
    } else {
      Stream.of(value.toString().split(","))
          .map(String::trim)
          .forEach(commands::add);
    }
    return commands.stream()
        .map(command -> Command.valueOf(command.toLowerCase()))
        .collect(Collectors.toList());
  }
}
